#import "XeeJPEGLoader.h"
#import "XeeBitmapImage.h"
#import "XeeYUVImage.h"
#import "XeeEXIFReader.h"



static void XeeJPEGErrorExit(j_common_ptr cinfo);
static void XeeJPEGEmitMessage(j_common_ptr cinfo,int msg_level);
static void XeeJPEGResetErrorMgr(j_common_ptr cinfo);

static void XeeDirectJPEGInitSource(j_decompress_ptr cinfo);
static boolean XeeDirectJPEGFillInputBuffer(j_decompress_ptr cinfo);
static void XeeDirectJPEGSkipInputData(j_decompress_ptr cinfo,long num);
static void XeeDirectJPEGTermSource(j_decompress_ptr cinfo);

static void XeeJPEGPlanarToChunky(uint8 *row,uint8 *y_row,uint8 *cb_row,uint8 *cr_row,int width);



@implementation XeeJPEGImage

+(NSArray *)fileTypes
{
	return [NSArray arrayWithObjects:@"jpg",@"jpeg",@"jpe",@"'JPEG'",nil];
}

+(BOOL)canOpenFile:(NSString *)name firstBlock:(NSData *)block attributes:(NSDictionary *)attributes;
{
	const unsigned char *head=[block bytes];
	int len=[block length];

	if(len>=2&&head[0]==0xff&&head[1]==0xd8) return YES;

	return NO;
}

-(SEL)initLoader
{
	jpeg_created=NO;
	ycbcr_buffers=NULL;
	thumb_ptr=NULL;
	thumb_len=0;

	jpeg_std_error(&jerr);
	jerr.error_exit=XeeJPEGErrorExit;
	jerr.emit_message=XeeJPEGEmitMessage;
	jerr.reset_error_mgr=XeeJPEGResetErrorMgr;
	cinfo.err=&jerr;

	jpeg_create_decompress(&cinfo);
	jpeg_created=YES;

	cinfo.dct_method=JDCT_IFAST;

	jpeg_save_markers(&cinfo,JPEG_APP0+1,0xffff);
	jpeg_save_markers(&cinfo,JPEG_COM,0xffff);

	jpeg_stdio_src(&cinfo,[[self fileHandle] filePointer]);
	jpeg_read_header(&cinfo,TRUE);

	width=cinfo.image_width;
	height=cinfo.image_height;

	switch(cinfo.jpeg_color_space)
	{
		case JCS_GRAYSCALE: [self setDepthGrey:8]; break;
		case JCS_RGB: [self setDepthRGB:8]; break;
		case JCS_YCbCr:
			[self setDepth:[NSString stringWithFormat:
			@"YCbCr H%dV%d",cinfo.max_h_samp_factor,cinfo.max_v_samp_factor]
			iconName:@"depth_rgb"];
		break;
		case JCS_CMYK: [self setDepthCMYK:8 alpha:NO]; break;
		case JCS_YCCK: [self setDepth:@"YCCK"]; break;
		default: [self setDepth:@"Unknown"]; break;
	}

	NSMutableArray *comments=nil;

	for(struct jpeg_marker_struct *marker=cinfo.marker_list;marker;marker=marker->next)
	{
		if(marker->marker==JPEG_APP0+1&&marker->data_length>6&&
		marker->data[0]=='E'&&marker->data[1]=='x'&&marker->data[2]=='i'&&
		marker->data[3]=='f'&&marker->data[4]==0&&marker->data[5]==0)
		{
			XeeEXIFReader *exif=[[XeeEXIFReader alloc] initWithBuffer:marker->data length:marker->data_length];

			[self setCorrectOrientation:[exif integerForTag:XeeOrientationTag set:XeeStandardTagSet]];

			int thumb_offs=[exif integerForTag:XeeThumbnailOffsetTag set:XeeStandardTagSet];
			if(thumb_offs)
			{
				thumb_len=[exif integerForTag:XeeThumbnailLengthTag set:XeeStandardTagSet];
				thumb_ptr=marker->data+6+thumb_offs;
			}

			[properties addObjectsFromArray:[exif propertyArray]];
		}
		else if(marker->marker==JPEG_COM)
		{
			if(!comments)
			{
				comments=[NSMutableArray array];
				[properties addObject:[XeePropertyItem itemWithLabel:
				NSLocalizedString(@"File comments",@"File comments section title")
				value:comments]];
			}
			[comments addObject:[XeePropertyItem itemWithLabel:@""
			value:[[[NSString alloc] initWithBytes:marker->data length:marker->data_length
			encoding:NSISOLatin1StringEncoding] autorelease]]];
		}
	}

	[self setFormat:@"JPEG"];

	if(thumbonly&&thumb_ptr) return @selector(loadThumbnail);
	return @selector(startLoading);
}

-(void)deallocLoader
{
	if(jpeg_created) jpeg_destroy_decompress(&cinfo);
	free(ycbcr_buffers);
}

-(SEL)startLoading
{
	if([[NSUserDefaults standardUserDefaults] boolForKey:@"jpegYUV"]
	&&cinfo.jpeg_color_space==JCS_YCbCr&&cinfo.comp_info[0].h_samp_factor==2
	&&(cinfo.comp_info[0].v_samp_factor==2||cinfo.comp_info[0].v_samp_factor==1)
	&&cinfo.comp_info[1].h_samp_factor==1&&cinfo.comp_info[1].v_samp_factor==1
	&&cinfo.comp_info[2].h_samp_factor==1&&cinfo.comp_info[2].v_samp_factor==1)
	{
		mainimage=[[[XeeYUVImage alloc] initWithWidth:width height:height] autorelease];
		cinfo.raw_data_out=TRUE;

		int y_width=(width+7)&~7;
		int y_rows=8*cinfo.comp_info[0].v_samp_factor;
		int cbcr_width=(width/2+7)&~7;

		ycbcr_buffers=malloc(y_rows*y_width+2*8*cbcr_width);
		if(!ycbcr_buffers) return NULL;

		y_buf=ycbcr_buffers;
		cb_buf=ycbcr_buffers+y_rows*y_width;
		cr_buf=ycbcr_buffers+y_rows*y_width+8*cbcr_width;

		for(int i=0;i<y_rows;i++) y_lines[i]=y_buf+i*y_width;
		for(int i=0;i<8;i++) cb_lines[i]=cb_buf+i*cbcr_width;
		for(int i=0;i<8;i++) cr_lines[i]=cr_buf+i*cbcr_width;
		image[0]=y_lines;
		image[1]=cb_lines;
		image[2]=cr_lines;
	}
	else
	{
		int type;
		if(cinfo.jpeg_color_space==JCS_GRAYSCALE) type=XeeBitmapTypeLuma8;
		else type=XeeBitmapTypeRGB8;

		mainimage=[[[XeeBitmapImage alloc] initWithType:type width:width height:height] autorelease];
	}

	if(!mainimage) return NULL;

	[mainimage setDepth:[self depth]];
	[mainimage setDepthIcon:[self depthIcon]];
	if(correctorientation) [mainimage setCorrectOrientation:correctorientation];
	[self addSubImage:mainimage];

	jpeg_start_decompress(&cinfo);

	if(cinfo.raw_data_out) return @selector(loadYUV);
	else return @selector(loadRGB);
}


-(SEL)loadYUV
{
	unsigned char *maindata=[mainimage data];
	int bprow=[mainimage bytesPerRow];

	int start_line=cinfo.output_scanline;
	int num_lines=8*cinfo.comp_info[0].v_samp_factor;
	jpeg_read_raw_data(&cinfo,image,2*DCTSIZE);

	for(int y=0;y<num_lines;y++)
	{
		if(start_line+y>=height) break;
		unsigned char *row=maindata+(start_line+y)*bprow;

		JSAMPLE *y_row=y_lines[y];
		JSAMPLE *cb_row=cb_lines[y/cinfo.comp_info[0].v_samp_factor];
		JSAMPLE *cr_row=cr_lines[y/cinfo.comp_info[0].v_samp_factor];

		XeeJPEGPlanarToChunky(row,y_row,cb_row,cr_row,width);
	}

	[mainimage setCompletedRowCount:cinfo.output_scanline];

	if(cinfo.output_scanline>=cinfo.output_height)
	{
		loaded=YES;
		return @selector(loadThumbnail);
	}
	return @selector(loadYUV);
}

-(SEL)loadRGB
{
	unsigned char *maindata=[mainimage data];
	int bprow=[mainimage bytesPerRow];

	for(int i=0;i<16;i++)
	{
		unsigned char *row=maindata+cinfo.output_scanline*bprow;
		jpeg_read_scanlines(&cinfo,&row,1);
		[mainimage setCompletedRowCount:cinfo.output_scanline];

		if(cinfo.output_scanline>=cinfo.output_height)
		{
			loaded=YES;
			return @selector(loadThumbnail);
		}
	}
	return @selector(loadRGB);
}

-(SEL)loadThumbnail
{
	if(thumb_ptr)
	{
		XeeImage *thumbnail=[[XeeDirectJPEGImage alloc] initWithBytes:thumb_ptr length:thumb_len];
		if(thumbnail)
		{
			if(correctorientation) [thumbnail setCorrectOrientation:correctorientation];
			[self addSubImage:thumbnail];
			[thumbnail release];

			loaded=YES; // for thumbnail-only loading
		}
	}
	return NULL;
}

-(int)losslessSaveFlags
{
	return 0;
}

-(BOOL)losslessSaveTo:(NSString *)path flags:(int)flags
{
	return NO;
}

@end



@implementation XeeBrokenJPEGImage

+(BOOL)canOpenFile:(NSString *)name firstBlock:(NSData *)block attributes:(NSDictionary *)attributes;
{
	const unsigned char *head=[block bytes];
	int len=[block length];

	NSString *ext=[[name pathExtension] lowercaseString];
	if([ext isEqual:@"jpg"]||[ext isEqual:@"jpeg"]||[ext isEqual:@"jpe"]||[attributes fileHFSTypeCode]=='JPEG')
	{
		for(int i=1;i<len-1;i++) if(head[i]==0xff&&head[i+1]==0xd8) return YES;
	}

	return NO;
}

-(SEL)initLoader
{
	XeeFileHandle *fh=[self fileHandle];
	uint8 prev=0;
	for(;;)
	{
		uint8 b=[fh readUInt8];
		if(prev==0xff&&b==0xd8) break;
		prev=b;
	}
	FILE *fp=[fh filePointer];
	ungetc(0xd8,fp);
	ungetc(0xff,fp);

	return [super initLoader];
}

@end


@implementation XeeDirectJPEGImage

-(id)initWithBytes:(const void *)bytes length:(int)len
{
	if(self=[super init])
	{
		struct jpeg_decompress_struct cinfo; 
		struct jpeg_error_mgr jerr; 
		struct jpeg_source_mgr src;

		jpeg_std_error(&jerr);
		jerr.error_exit=XeeJPEGErrorExit;
		jerr.emit_message=XeeJPEGEmitMessage;
		jerr.reset_error_mgr=XeeJPEGResetErrorMgr;
		cinfo.err=&jerr;

		@try
		{
			jpeg_create_decompress(&cinfo);

		src.init_source=XeeDirectJPEGInitSource;
		src.fill_input_buffer=XeeDirectJPEGFillInputBuffer;
		src.skip_input_data=XeeDirectJPEGSkipInputData;
		src.resync_to_restart=jpeg_resync_to_restart;
		src.term_source=XeeDirectJPEGTermSource;
		src.bytes_in_buffer=len;
		src.next_input_byte=bytes;
		cinfo.src=&src;

			cinfo.dct_method=JDCT_IFAST;
			jpeg_read_header(&cinfo,TRUE);

			int type;
			if(cinfo.jpeg_color_space==JCS_GRAYSCALE) type=XeeBitmapTypeLuma8;
			else type=XeeBitmapTypeRGB8;

			if(![self allocWithType:type width:cinfo.image_width height:cinfo.image_height])
			@throw @"Out of memory";

			switch(cinfo.jpeg_color_space)
			{
				case JCS_GRAYSCALE: [self setDepthGrey:8]; break;
				case JCS_RGB: [self setDepthRGB:8]; break;
				case JCS_YCbCr:
					[self setDepth:[NSString stringWithFormat:
					@"YCbCr H%dV%d",cinfo.max_h_samp_factor,cinfo.max_v_samp_factor]
					iconName:@"depth_rgb"];
				break;
				case JCS_CMYK: [self setDepthCMYK:8 alpha:NO]; break;
				case JCS_YCCK: [self setDepth:@"YCCK"]; break;
				default: [self setDepth:@"Unknown"]; break;
			}

			[self setFormat:@"JPEG"];

			jpeg_start_decompress(&cinfo);
			for(int i=0;i<height;i++)
			{
				uint8 *row=data+cinfo.output_scanline*bytesperrow;
				jpeg_read_scanlines(&cinfo,&row,1);
			}

			[self setCompleted];

			jpeg_destroy_decompress(&cinfo);
		}
		@catch(id e)
		{
			NSLog(@"XeeDirectJPEGImage creation error: %@",e);
			jpeg_destroy_decompress(&cinfo);
			[self release];
			return nil;
		}
	}
	return self;
}

-(id)initWithData:(NSData *)jpegdata
{
	if(!jpegdata)
	{
		[self release];
		return nil;
	}
	return [self initWithBytes:[jpegdata bytes] length:[jpegdata length]];
}
 
@end



static void XeeJPEGErrorExit(j_common_ptr cinfo)
{
	char buffer[JMSG_LENGTH_MAX];
	(*cinfo->err->format_message)(cinfo,buffer);
	[NSException raise:@"XeeLibJPEGException" format:@"libjpeg error: %s",buffer];
}

static void XeeJPEGEmitMessage(j_common_ptr cinfo,int msg_level) // Ignore warnings for now
{
}

static void XeeJPEGResetErrorMgr(j_common_ptr cinfo)
{
}

static void XeeDirectJPEGInitSource(j_decompress_ptr cinfo) {}

static boolean XeeDirectJPEGFillInputBuffer(j_decompress_ptr cinfo)
{
NSLog(@"fillinputbuffer %d %x",cinfo->src->bytes_in_buffer,cinfo->src->next_input_byte);
	static uint8 eoi[2]={0xff,0xd9};
	cinfo->src->next_input_byte=eoi;
	cinfo->src->bytes_in_buffer=2;
	return TRUE;
}

static void XeeDirectJPEGSkipInputData(j_decompress_ptr cinfo,long num)
{
	if(num>0)
	{
		cinfo->src->next_input_byte+=num;
		cinfo->src->bytes_in_buffer-=num;
	}
}

static void XeeDirectJPEGTermSource(j_decompress_ptr cinfo) {}

static void XeeJPEGPlanarToChunky(uint8 *row,uint8 *y_row,uint8 *cb_row,uint8 *cr_row,int width)
{
	#ifdef __BIG_ENDIAN__
	#define MOVE(val,from,to) ( ( ((to)>(from)) ? ((val)>>(((to)-(from))*8)) : (((val)<<((from)-(to))*8)) ) & (0xff<<(24-(to)*8)) )
	#else
	#define MOVE(val,from,to) ( ( ((to)>(from)) ? ((val)<<(((to)-(from))*8)) : (((val)>>((from)-(to))*8)) ) & (0xff<<((to)*8)) )
	#endif

	int n=(width+1)/4;
	uint32 *row_l=(uint32 *)row;
	uint32 *y_l=(uint32 *)y_row;
	uint32 *cb_l=(uint32 *)cb_row;
	uint32 *cr_l=(uint32 *)cr_row;
	uint32 y,cb,cr;

	for(;;)
	{
		if(!n--) break;

		y=*y_l++;
		cb=*cb_l++;
		cr=*cr_l++;
		*row_l++=MOVE(cb,0,0)|MOVE(y,0,1)|MOVE(cr,0,2)|MOVE(y,1,3);
		*row_l++=MOVE(cb,1,0)|MOVE(y,2,1)|MOVE(cr,1,2)|MOVE(y,3,3);

		if(!n--) break;

		y=*y_l++;
		*row_l++=MOVE(cb,2,0)|MOVE(y,0,1)|MOVE(cr,2,2)|MOVE(y,1,3);
		*row_l++=MOVE(cb,3,0)|MOVE(y,2,1)|MOVE(cr,3,2)|MOVE(y,3,3);
	}

	if((width+1)&2)
	{
		uint8 *row_b=(uint8 *)row_l;
		uint8 *y_b=(uint8 *)y_l;
		uint8 *cb_b=(uint8 *)cb_l;
		uint8 *cr_b=(uint8 *)cr_l;
		*row_b++=*cb_b;
		*row_b++=*y_b++;
		*row_b++=*cr_b;
		*row_b=*y_b;
	}
}

static const unsigned char gammatable[256]=
{
	0x00,0x00,0x01,0x01,0x02,0x02,0x03,0x03,0x04,0x04,0x05,0x05,0x06,0x07,0x07,0x08,
	0x09,0x09,0x0a,0x0b,0x0b,0x0c,0x0d,0x0d,0x0e,0x0f,0x10,0x10,0x11,0x12,0x13,0x13,
	0x14,0x15,0x16,0x17,0x17,0x18,0x19,0x1a,0x1b,0x1b,0x1c,0x1d,0x1e,0x1f,0x1f,0x20,
	0x21,0x22,0x23,0x24,0x25,0x25,0x26,0x27,0x28,0x29,0x2a,0x2b,0x2c,0x2c,0x2d,0x2e,
	0x2f,0x30,0x31,0x32,0x33,0x34,0x35,0x35,0x36,0x37,0x38,0x39,0x3a,0x3b,0x3c,0x3d,
	0x3e,0x3f,0x40,0x41,0x42,0x43,0x44,0x45,0x45,0x46,0x47,0x48,0x49,0x4a,0x4b,0x4c,
	0x4d,0x4e,0x4f,0x50,0x51,0x52,0x53,0x54,0x55,0x56,0x57,0x58,0x59,0x5a,0x5b,0x5c,
	0x5d,0x5e,0x5f,0x60,0x61,0x62,0x63,0x64,0x65,0x67,0x68,0x69,0x6a,0x6b,0x6c,0x6d,
	0x6e,0x6f,0x70,0x71,0x72,0x73,0x74,0x75,0x76,0x77,0x78,0x79,0x7b,0x7c,0x7d,0x7e,
	0x7f,0x80,0x81,0x82,0x83,0x84,0x85,0x86,0x87,0x89,0x8a,0x8b,0x8c,0x8d,0x8e,0x8f,
	0x90,0x91,0x92,0x94,0x95,0x96,0x97,0x98,0x99,0x9a,0x9b,0x9c,0x9e,0x9f,0xa0,0xa1,
	0xa2,0xa3,0xa4,0xa5,0xa7,0xa8,0xa9,0xaa,0xab,0xac,0xad,0xaf,0xb0,0xb1,0xb2,0xb3,
	0xb4,0xb5,0xb7,0xb8,0xb9,0xba,0xbb,0xbc,0xbd,0xbf,0xc0,0xc1,0xc2,0xc3,0xc4,0xc6,
	0xc7,0xc8,0xc9,0xca,0xcb,0xcd,0xce,0xcf,0xd0,0xd1,0xd3,0xd4,0xd5,0xd6,0xd7,0xd8,
	0xda,0xdb,0xdc,0xdd,0xde,0xe0,0xe1,0xe2,0xe3,0xe4,0xe6,0xe7,0xe8,0xe9,0xea,0xec,
	0xed,0xee,0xef,0xf0,0xf2,0xf3,0xf4,0xf5,0xf6,0xf8,0xf9,0xfa,0xfb,0xfd,0xfe,0xff,
};
